<?php

namespace Drupal\views_combine;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\views\ViewExecutable;

/**
 * Provides views query combiner.
 */
class ViewsCombiner {

  /**
   * The base view.
   *
   * @var \Drupal\views\ViewExecutable
   */
  protected $view;

  /**
   * The base view identifier.
   *
   * @var string
   */
  protected $viewId;

  /**
   * An array of view queries to combine.
   *
   * @var \Drupal\Core\Database\Query\Select[]
   */
  protected $queries = [];

  /**
   * An array of query fields and expressions.
   *
   * @var array
   */
  protected $fields = [];

  /**
   * An array of query order fields and keys.
   *
   * @var array
   */
  protected $orders = [];

  /**
   * An array of query order key directions.
   *
   * @var array
   */
  protected $orderDirections = [];

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * ViewsCombiner constructor.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(Connection $connection, EntityTypeManagerInterface $entityTypeManager) {
    $this->connection = $connection;
    $this->entityTypeManager = $entityTypeManager;
  }

  /**
   * Combine view queries.
   *
   * @var \Drupal\views\ViewExecutable $view
   *   The base view.
   *
   * @see views_combine_views_post_build()
   */
  public function combine(ViewExecutable $view) {
    $this->view = $view;
    $this->viewId = $this->view->id() . ':' . $this->view->current_display;
    $this->setViews();
    if (count($this->queries) > 1) {
      // Establish collection of query fields and orders.
      foreach ($this->queries as $query) {
        $this->fields += $query->getFields() + $query->getExpressions();
        $order_key = 0;
        foreach (array_keys($query->getOrderBy()) as $order_field) {
          $this->orders[$order_field] = $order_key;
          $order_key++;
        }
      }
      $this->setUnions();
    }
  }

  /**
   * Build view with inheritance from base view.
   *
   * @param string $view_id
   *   The view identifier.
   * @param string $display_id
   *   The display identifier.
   * @param \Drupal\views\ViewExecutable $base_view
   *   The base view.
   *
   * @return \Drupal\views\ViewExecutable|null
   *   Returns the built view.
   */
  public function buildView(string $view_id, string $display_id, ViewExecutable $base_view) {
    $view_exec = NULL;
    try {
      /** @var \Drupal\views\Entity\View $view */
      $view = $this->entityTypeManager->getStorage('view')
        ->load($view_id);

      if ($view && isset($view->get('display')[$display_id])) {
        $view_exec = $view->getExecutable();
        $view_exec->setDisplay($display_id);
        $view_exec->setExposedInput($base_view->exposed_raw_input);
        $view_exec->preExecute($base_view->args);
        $view_exec->build();
      }
    }
    catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
      // Unable to load view storage.
    }
    return $view_exec;
  }

  /**
   * Set view queries to combine.
   */
  protected function setViews() {
    /** @var \Drupal\Core\Database\Query\Select $query */
    $query = $this->view->getQuery()->query();
    $query->addMetaData('view_id', $this->viewId);
    $this->queries[] = $query;

    foreach ($this->view->field as $field_handler) {
      if ($field_handler->realField === 'views_combine' && $field_handler->options['view_id']) {
        [$view_id, $display_id] = explode(':', $field_handler->options['view_id']);

        if ($view = $this->buildView($view_id, $display_id, $this->view)) {
          /** @var \Drupal\Core\Database\Query\Select $query */
          $query = $view->getQuery()->query();
          $query->addMetaData('view_id', $view->id() . ':' . $view->current_display);
          $this->queries[] = $query;
        }
      }
    }
  }

  /**
   * Set views query fields.
   *
   * Each UNION query must have matching column names and positions. Otherwise,
   * the query execution throws a fatal or may have incorrect data results. Plus
   * normalizing order field names presents the capability for robust sorting.
   */
  protected function setFields() {
    foreach ($this->queries as $key => &$query) {
      $fields =& $query->getFields();
      $expressions =& $query->getExpressions();

      // Convert fields to expressions to manage column positions.
      foreach ($fields as $schema) {
        $query->addExpression($this->connection->escapeField($schema['table'] . '.' . $schema['field']), $schema['alias']);
      }
      $fields = [];

      // Normalize expression values and column positions.
      foreach (array_diff_key($this->fields, $expressions) as $schema) {
        $query->addExpression('NULL', $schema['alias']);
      }
      $expressions = array_merge(array_flip(array_keys($this->fields)), $expressions);

      // Set order fields. Order direction prefers the topmost view.
      $orders =& $query->getOrderBy();
      foreach ($orders as $field => $direction) {
        if (!isset($this->orderDirections[$this->orders[$field]])) {
          $this->orderDirections[$this->orders[$field]] = $direction;
        }
        $query->addExpression($this->connection->escapeField($expressions[$field]['expression']), "_order_{$this->orders[$field]}");
      }
      for ($i = count($orders); $i < count(array_unique($this->orders)); $i++) {
        $query->addExpression('NULL', "_order_$i");
      }
      $orders = [];

      // Add view display identifier.
      $query->addExpression(":view_$key", '_view_id', [
        ":view_$key" => $query->getMetaData('view_id'),
      ]);
    }
  }

  /**
   * Set views query filters.
   *
   * Specifically, the views_combine exposed filter(s). Sets an impossible
   * condition to exclude results from a particular view display.
   *
   * @see \Drupal\views_combine\Plugin\views\filter\Combine
   */
  protected function setFilters() {
    $filters = [];
    foreach ($this->view->filter as $filter) {
      if ($filter->realField === 'views_combine' && $filter->value) {
        $filters[$filter->operator] = array_merge($filters, $filter->value);
      }
    }
    foreach ($this->queries as &$query) {
      $view_id = $query->getMetaData('view_id');
      foreach ($filters as $operator => $value) {
        $in_value = in_array($view_id, $value) || (in_array('current_view', $value) && $view_id === $this->viewId);
        if (($operator === 'in' && !$in_value) || ($operator === 'not in' && $in_value)) {
          $query->where('0 = 1');
        }
      }
    }
  }

  /**
   * Set views query unions.
   *
   * First query is presumed to be the base view query. All proceeding queries
   * are unionized under the base query.
   */
  protected function setUnions() {
    $this->setFields();
    $this->setFilters();

    // Set base query order and unions.
    $base_query = array_shift($this->queries);
    foreach ($this->queries as $query) {
      $base_query->union($query);
    }
    foreach ($this->orders as $order_key) {
      $base_query->orderBy("_order_$order_key", $this->orderDirections[$order_key]);
    }
    $this->view->build_info['query'] = $base_query;

    // Set count query with distinct to prevent optimizations.
    if ($this->view->display_handler->usesPager()) {
      $count_query = clone $base_query;
      $count_query->distinct();
      $unions =& $count_query->getUnion();
      foreach ($unions as $union_query) {
        $union_query['query']->distinct();
      }
      $this->view->build_info['count_query'] = $count_query;
    }
  }

}
